/*
 * Copyright (c) 2018-2019,2021-2023 LAAS/CNRS
 * All rights reserved.
 *
 * Redistribution  and  use  in  source  and binary  forms,  with  or  without
 * modification, are permitted provided that the following conditions are met:
 *
 *   1. Redistributions of  source  code must retain the  above copyright
 *      notice and this list of conditions.
 *   2. Redistributions in binary form must reproduce the above copyright
 *      notice and  this list of  conditions in the  documentation and/or
 *      other materials provided with the distribution.
 *
 * THE SOFTWARE  IS PROVIDED "AS IS"  AND THE AUTHOR  DISCLAIMS ALL WARRANTIES
 * WITH  REGARD   TO  THIS  SOFTWARE  INCLUDING  ALL   IMPLIED  WARRANTIES  OF
 * MERCHANTABILITY AND  FITNESS.  IN NO EVENT  SHALL THE AUTHOR  BE LIABLE FOR
 * ANY  SPECIAL, DIRECT,  INDIRECT, OR  CONSEQUENTIAL DAMAGES  OR  ANY DAMAGES
 * WHATSOEVER  RESULTING FROM  LOSS OF  USE, DATA  OR PROFITS,  WHETHER  IN AN
 * ACTION OF CONTRACT, NEGLIGENCE OR  OTHER TORTIOUS ACTION, ARISING OUT OF OR
 * IN CONNECTION WITH THE USE OR PERFORMANCE OF THIS SOFTWARE.
 *
 *                                           Anthony Mallet on Fri Jun  1 2018
 */
#include "acuavpos.h"

#include <sys/time.h>
#include <aio.h>
#include <err.h>
#include <unistd.h>

#include <algorithm>
#include <cmath>
#include <cstdio>

#include "Eigen/Core"
#include "Eigen/Dense"

#include "codels.h"

/*
 * --- uavpos_controller ---------------------------------------------------
 *
 * Implements the position bits of the controller described in:
 *
 * T. Lee, M. Leoky and N. H. McClamroch, "Geometric tracking control of a
 * quadrotor UAV on SE(3)", 49th IEEE Conference on Decision and Control
 * (CDC), Atlanta, GA, 2010, pp. 5420-5425.
 */

int
uavpos_controller(const uavpos_ids_body_s *body,
                  const uavpos_ids_servo_s *servo,
                  const or_pose_estimator_state *state,
                  const or_rigid_body_state *desired,
                  uavpos_log_s *log,
                  or_uav_input *uav_input)
{
  using namespace Eigen;

  Matrix3d Rd;
  Quaternion<double> qd;
  Vector3d xd, vd, wd, ad, awd;

  Vector3d x, v;

  Vector3d ex, ev;
  static Vector3d Iex;

  Vector3d f;
  Vector3d b3d, k;
  double xythrust, theta;

  int i;

  /* gains */
  const Array3d Kp(servo->gain.Kpxy, servo->gain.Kpxy, servo->gain.Kpz);
  const Array3d Ki(servo->gain.Kixy, servo->gain.Kixy, servo->gain.Kiz);
  const Array3d Kv(servo->gain.Kvxy, servo->gain.Kvxy, servo->gain.Kvz);


  /* desired state - values are always valid due to uavpos_reference_check() */
  xd <<
    desired->pos._value.x, desired->pos._value.y, desired->pos._value.z;
  qd.coeffs() <<
    desired->att._value.qx, desired->att._value.qy, desired->att._value.qz,
    desired->att._value.qw;

  vd <<
    desired->vel._value.vx, desired->vel._value.vy, desired->vel._value.vz;
  wd <<
    desired->avel._value.wx, desired->avel._value.wy, desired->avel._value.wz;

  ad <<
    desired->acc._value.ax, desired->acc._value.ay, desired->acc._value.az;
  awd <<
    desired->aacc._value.awx,
    desired->aacc._value.awy,
    desired->aacc._value.awz;

  if (desired->intrinsic) {
    vd = qd * vd;
    wd = qd * wd;
    ad = qd * ad;
    awd = qd * awd;
  }


  /* current state */
  if (state->pos._present) {
    x << state->pos._value.x, state->pos._value.y, state->pos._value.z;
  } else {
    x = xd;
    Iex << 0., 0., 0.;
    ad = Eigen::Vector3d(0, 0, - servo->emerg.descent);
  }

  if (state->vel._present) {
    v << state->vel._value.vx, state->vel._value.vy, state->vel._value.vz;
  } else {
    v = vd;
    ad = Eigen::Vector3d(0, 0, - servo->emerg.descent);
  }


  /* position error */
  ex = x - xd;
  for(i = 0; i < 3; i++)
    if (fabs(ex(i)) > servo->sat.x) ex(i) = copysign(servo->sat.x, ex(i));

  Iex += ex * uavpos_control_period_ms/1000.;
  for(i = 0; i < 3; i++)
    if (fabs(Iex(i)) > servo->sat.ix) Iex(i) = copysign(servo->sat.ix, Iex(i));


  /* velocity error */
  ev = v - vd;
  for(i = 0; i < 3; i++)
    if (fabs(ev(i)) > servo->sat.v) ev(i) = copysign(servo->sat.v, ev(i));


  /* desired thrust */
  f =
    - Kp * ex.array() - Kv * ev.array() - Ki * Iex.array()
    + body->mass * (Eigen::Vector3d(0, 0, 9.81) + ad).array();

  /* desired orientation */
  b3d = qd.matrix().col(2).normalized();

  /* make desired orientation admissible - this is a specialized version of the
   * algorithm presented in Antonio Franchi, Ruggero Carli, Davide Bicego,
   * Markus Ryll. Full-Pose Tracking Control for Aerial Robotic Systems With
   * Laterally Bounded Input Force. IEEE Transactions on Robotics, IEEE, 2018,
   * 34 (2), pp.534-541.
   * Here, a cylindric approximation is used. */

  k = f.cross(b3d);
  xythrust = k.norm();
  if (xythrust > body->rxy) {
    /* an admissible orientation is such that the lateral forces are equal to
     * rxy. Theta is the maximum admissible angle in the (f, b3d) plane from
     * f to b3d. */
    theta = std::asin(body->rxy / f.norm());

    /* unit vector perpendicular to the (f, b3d) plane */
    k /= xythrust;

    /* rotate f by theta around k and get new desired orientation */
    b3d = (f * std::cos(theta) + k.cross(f) * std::sin(theta)).normalized();
  }

  Rd.col(2) = b3d;
  Rd.col(1) = Rd.col(2).cross(qd.matrix().col(0)).normalized();
  Rd.col(0) = Rd.col(1).cross(Rd.col(2));

  qd = Quaternion<double>(Rd);


  /* output */
  uav_input->intrinsic = false;

  uav_input->thrust._present = true;
  uav_input->thrust._value.x = f(0);
  uav_input->thrust._value.y = f(1);
  uav_input->thrust._value.z = f(2);

  uav_input->att._present = true;
  uav_input->att._value.qw = qd.w();
  uav_input->att._value.qx = qd.vec()(0);
  uav_input->att._value.qy = qd.vec()(1);
  uav_input->att._value.qz = qd.vec()(2);

  uav_input->avel._present = true;
  uav_input->avel._value.wx = wd(0);
  uav_input->avel._value.wy = wd(1);
  uav_input->avel._value.wz = wd(2);

  uav_input->aacc._present = true;
  uav_input->aacc._value.awx = awd(0);
  uav_input->aacc._value.awy = awd(1);
  uav_input->aacc._value.awz = awd(2);


  /* logging */
  if (log->req.aio_fildes >= 0) {
    log->total++;
    if (log->total % log->decimation == 0) {
      if (log->pending) {
        if (aio_error(&log->req) != EINPROGRESS) {
          log->pending = false;
          if (aio_return(&log->req) <= 0) {
            warn("log");
            close(log->req.aio_fildes);
            log->req.aio_fildes = -1;
          }
        } else {
          log->skipped = true;
          log->missed++;
        }
      }
    }

    if (log->req.aio_fildes >= 0 && !log->pending) {
      double yaw = atan2(Rd(1,0), Rd(0,0));
      double pitch = asin(-Rd(2,0));
      double roll = atan2(Rd(2,1), Rd(2,2));

      log->req.aio_nbytes = snprintf(
        log->buffer, sizeof(log->buffer),
        "%s" uavpos_log_fmt "\n",
        log->skipped ? "\n" : "",
        uav_input->ts.sec, uav_input->ts.nsec,
        uav_input->ts.sec - state->ts.sec +
        (uav_input->ts.nsec - state->ts.nsec)*1e-9,
        f(0), f(1), f(2),
        xd(0), xd(1), xd(2), roll, pitch, yaw,
        vd(0), vd(1), vd(2), wd(0), wd(1), wd(2),
        ad(0), ad(1), ad(2), awd(0), awd(1), awd(2),
        ex(0), ex(1), ex(2), ev(0), ev(1), ev(2));

      if (aio_write(&log->req)) {
        warn("log");
        close(log->req.aio_fildes);
        log->req.aio_fildes = -1;
      } else
        log->pending = true;

      log->skipped = false;
    }
  }

  return 0;
}


/*
 * --- uavpos_state_check --------------------------------------------------
 *
 * Return non-zero in case of emergency
 */

int
uavpos_state_check(const struct timeval now,
                   const uavpos_ids_servo_s *servo,
                   or_pose_estimator_state *state)
{
  int e = UAV_EOK;

  /* check state */
  if (now.tv_sec + 1e-6 * now.tv_usec >
      0.5 + state->ts.sec + 1e-9 * state->ts.nsec) {
    state->pos._present = false;
    state->att._present = false;
    state->vel._present = false;
    state->avel._present = false;
    return UAV_ETS;
  }

  if (!state->pos._present
      || std::isnan(state->pos._value.x)
      || !state->pos_cov._present
      || state->pos_cov._value.cov[0] > servo->emerg.dx
      || state->pos_cov._value.cov[2] > servo->emerg.dx
      || state->pos_cov._value.cov[5] > servo->emerg.dx) {
    state->pos._present = false;
    e |= UAV_EPOS;
  }

  if (!state->vel._present
      || std::isnan(state->vel._value.vx)
      || !state->vel_cov._present
      || state->vel_cov._value.cov[0] > servo->emerg.dv
      || state->vel_cov._value.cov[2] > servo->emerg.dv
      || state->vel_cov._value.cov[5] > servo->emerg.dv) {
    state->vel._present = false;
    e |= UAV_EVEL;
  }

  return e;
}


/*
 * --- uavpos_reference_check ----------------------------------------------
 *
 * Update missing fields of the reference by integrating other fields
 */

void
uavpos_reference_check(const struct timeval now,
                       or_rigid_body_state *reference)
{
  static const double dt = uavpos_control_period_ms / 1000.;
  static const double dt2 = dt * dt;
  static const double dt2_2 = dt2 / 2.;

  or_t3d_pos *p = &reference->pos._value;
  or_t3d_att *q = &reference->att._value;
  or_t3d_vel *v = &reference->vel._value;
  or_t3d_avel *w = &reference->avel._value;
  or_t3d_acc *a = &reference->acc._value;
  or_t3d_aacc *aw = &reference->aacc._value;

  /* deal with obsolete reference */
  if (now.tv_sec + 1e-6 * now.tv_usec >
      0.5 + reference->ts.sec + 1e-9 * reference->ts.nsec) {
    reference->vel._present = true;
    v->vx = v->vy = v->vz = 0.;

    reference->avel._present = true;
    w->wx = w->wy = w->wz = 0.;

    reference->acc._present = true;
    a->ax = a->ay = a->az = 0.;
  }

  if (!reference->pos._present) {
    /* use previous pos and integrate vel, acc */
    Eigen::Vector3d dp(dt * v->vx + dt2_2 * a->ax,
                       dt * v->vy + dt2_2 * a->ay,
                       dt * v->vz + dt2_2 * a->az);
    if (reference->intrinsic)
      dp = Eigen::Quaternion<double>(q->qw, q->qx, q->qy, q->qz) * dp;

    p->x += dp(0);
    p->y += dp(1);
    p->z += dp(2);
  }

  if (!reference->vel._present) {
    /* use previous vel and integrate acc */
    v->vx += dt * a->ax;
    v->vy += dt * a->ay;
    v->vz += dt * a->az;
  }

  if (!reference->acc._present) {
    /* reset */
    a->ax = a->ay = a->az = 0.;
  }

  if (!reference->att._present) {
    /* use previous att and integrate avel */
    double a2 = dt2 * (w->wx * w->wx + w->wy * w->wy + w->wz * w->wz);

    if (a2 < 1e-1 /* otherwise do nothing, too high */) {
      Eigen::Quaternion<double> Q(q->qw, q->qx, q->qy, q->qz), dq;
      Eigen::Vector3d W(w->wx, w->wy, w->wz);
      if (reference->intrinsic)
        W = Q * W;

      dq.w() = 1 - a2/8; /* cos(a/2) */
      dq.vec() = dt * (0.5 - a2/48) /* dt * sin(a/2)/a */ * W;
      Q = dq * Q;

      q->qw = Q.w(); q->qx = Q.x(); q->qy = Q.y(); q->qz = Q.z();
    }
  }

  if (!reference->avel._present) {
    /* reset */
    w->wx = w->wy = w->wz = 0.;
  }

  if (!reference->aacc._present) {
    /* reset */
    aw->awx = aw->awy = aw->awz = 0.;
  }
}
