/*
 * Copyright (c) 2015-2018,2021-2023 LAAS/CNRS
 * All rights reserved.
 *
 * Redistribution and use  in source  and binary  forms,  with or without
 * modification, are permitted provided that the following conditions are
 * met:
 *
 *   1. Redistributions of  source  code must retain the  above copyright
 *      notice and this list of conditions.
 *   2. Redistributions in binary form must reproduce the above copyright
 *      notice and  this list of  conditions in the  documentation and/or
 *      other materials provided with the distribution.
 *
 *                                      Anthony Mallet on Thu May 31 2018
 */
#include "acuavpos.h"

#include <sys/time.h>
#include <err.h>
#include <math.h>
#include <stdio.h>
#include <stdlib.h>

#include "uavpos_c_types.h"
#include "codels.h"


/* --- Task main -------------------------------------------------------- */


/** Codel uavpos_main_start of task main.
 *
 * Triggered by uavpos_start.
 * Yields to uavpos_init.
 */
genom_event
uavpos_main_start(uavpos_ids *ids, const uavpos_uav_input *uav_input,
                  const genom_context self)
{
  or_uav_input *input_data;

  ids->body = (uavpos_ids_body_s){
    /* mikrokopter quadrotors defaults */
    .mass = 1.0,
    .rxy = 0.0,
  };

  ids->servo = (uavpos_ids_servo_s){
    .sat = { .x = 0.10, .v = 0.1, .ix = 0.10 },
    .gain = {
      .Kpxy = 14., .Kvxy = 7., .Kpz = 20., .Kvz = 10.,
      .Kixy = 0., .Kiz = 0.
    },

    .emerg = {
      .descent = .1,
      .dx = 0.05 * 0.05 /9.,
      .dv = 0.2 * 0.2 /9.,
    }
  };

  ids->reference = (or_rigid_body_state){
    .ts = { .sec = 0, .nsec = 0 },
    .intrinsic = false,
    .pos._present = false,
    .att._present = false, .att._value = { .qw = 1. },
    .vel._present = false,
    .avel._present = false,
    .acc._present = false,
    .aacc._present = false,
    .jerk._present = false,
    .snap._present = false,
  };

  input_data = uav_input->data(self);
  *input_data = (or_uav_input){
    .ts = { .sec = 0, .nsec = 0 },
    .intrinsic = false,
    .thrust = { ._present = false },
    .att = { ._present = false },
    .avel = { ._present = false },
    .aacc = { ._present = false }
  };
  uav_input->write(self);


  /* init logging */
  ids->log = malloc(sizeof(*ids->log));
  if (!ids->log) abort();
  *ids->log = (uavpos_log_s){
    .req = {
      .aio_fildes = -1,
      .aio_offset = 0,
      .aio_buf = ids->log->buffer,
      .aio_nbytes = 0,
      .aio_reqprio = 0,
      .aio_sigevent = { .sigev_notify = SIGEV_NONE },
      .aio_lio_opcode = LIO_NOP
    },
    .pending = false, .skipped = false,
    .decimation = 1, .missed = 0, .total = 0
  };

  return uavpos_init;
}


/** Codel uavpos_main_init of task main.
 *
 * Triggered by uavpos_init.
 * Yields to uavpos_pause_init, uavpos_pause_control.
 */
genom_event
uavpos_main_init(or_rigid_body_state *reference,
                 const uavpos_state *state,
                 const uavpos_uav_input *uav_input,
                 const genom_context self)
{
  or_pose_estimator_state *state_data;
  or_uav_input *input_data;
  struct timeval tv;

  /* reset output */
  input_data = uav_input->data(self);
  if (!input_data) return uavpos_pause_init;

  gettimeofday(&tv, NULL);
  input_data->ts.sec = tv.tv_sec;
  input_data->ts.nsec = tv.tv_usec * 1000;

  uav_input->write(self);

  /* switch to servo mode upon reception of the first valid position or
   * velocity. Ensure that state has valid pos/att and initialize any empty
   * pos/att reference to it. */
  while (reference->pos._present || reference->vel._present) {
    if (state->read(self)) break;
    state_data = state->data(self);
    if (!state_data) break;
    if (!state_data->pos._present) break;
    if (!state_data->att._present) break;

    if (!reference->pos._present)
      reference->pos._value = state_data->pos._value;
    if (!reference->att._present)
      reference->att._value = state_data->att._value;

    return uavpos_pause_control;
  }

  return uavpos_pause_init;
}


/** Codel uavpos_main_control of task main.
 *
 * Triggered by uavpos_control.
 * Yields to uavpos_pause_control, uavpos_emergency.
 */
genom_event
uavpos_main_control(const uavpos_ids_body_s *body,
                    uavpos_ids_servo_s *servo,
                    const uavpos_state *state,
                    or_rigid_body_state *reference, uavpos_log_s **log,
                    const uavpos_uav_input *uav_input,
                    const genom_context self)
{
  or_pose_estimator_state *state_data;
  or_uav_input *input_data = uav_input->data(self);
  struct timeval tv;
  int e;

  gettimeofday(&tv, NULL);

  /* read current state */
  if (state->read(self) || !(state_data = state->data(self))) {
    warnx("state unavailable");
    warnx("emergency descent");
    return uavpos_emergency;
  }

  /* check state */
  e = uavpos_state_check(tv, servo, state_data);
  if (e) {
    if (e & UAV_ETS) warnx("obsolete state");
    if (e & UAV_EPOS) warnx("uncertain position");
    if (e & UAV_EVEL) warnx("uncertain velocity");
    warnx("emergency descent");
    return uavpos_emergency;
  }

  /* check reference */
  uavpos_reference_check(tv, reference);

  /* position controller */
  uavpos_controller(body, servo, state_data, reference, *log, input_data);

  /* output */
  input_data->ts = state_data->ts;
  uav_input->write(self);
  return uavpos_pause_control;
}


/** Codel uavpos_main_emergency of task main.
 *
 * Triggered by uavpos_emergency.
 * Yields to uavpos_pause_emergency, uavpos_control.
 */
genom_event
uavpos_main_emergency(const uavpos_ids_body_s *body,
                      uavpos_ids_servo_s *servo,
                      const uavpos_state *state,
                      or_rigid_body_state *reference,
                      uavpos_log_s **log,
                      const uavpos_uav_input *uav_input,
                      const genom_context self)
{
  or_pose_estimator_state *state_data;
  or_uav_input *input_data = uav_input->data(self);
  struct timeval tv;
  int e;

  gettimeofday(&tv, NULL);

  /* read current state */
  if (state->read(self) || !(state_data = state->data(self))) {
    state_data = &(or_pose_estimator_state){
      /* data without any 'present' field except current timestamp for logs */
      .ts.sec = tv.tv_sec, .ts.nsec = 1000 * tv.tv_usec
    };
  }

  /* check state */
  e = uavpos_state_check(tv, servo, state_data);
  if (!e) {
    warnx("recovered from emergency");
    return uavpos_control;
  }

  /* check reference */
  uavpos_reference_check(tv, reference);

  /* position controller */
  uavpos_controller(body, servo, state_data, reference, *log, input_data);

  /* output */
  input_data->ts = state_data->ts;
  uav_input->write(self);
  return uavpos_pause_emergency;
}


/** Codel mk_main_stop of task main.
 *
 * Triggered by uavpos_stop.
 * Yields to uavpos_ether.
 */
genom_event
mk_main_stop(const uavpos_uav_input *uav_input,
             const genom_context self)
{
  or_uav_input *input_data;
  struct timeval tv;

  input_data = uav_input->data(self);
  if (!input_data) return uavpos_ether;

  gettimeofday(&tv, NULL);
  input_data->ts.sec = tv.tv_sec;
  input_data->ts.nsec = tv.tv_usec * 1000;
  input_data->intrinsic = false;

  input_data->thrust = (optional_or_rb3d_force){
    ._present = true,
    ._value = { .x = 0., .y = 0., .z = 0. }
  };
  input_data->att._present = false;
  input_data->avel._present = false;
  input_data->aacc._present = false;

  uav_input->write(self);

  return uavpos_ether;
}


/* --- Activity servo --------------------------------------------------- */

/** Codel uavpos_servo_loop of activity servo.
 *
 * Triggered by uavpos_start.
 * Yields to uavpos_pause_start, uavpos_ether.
 * Throws uavpos_e_input.
 */
genom_event
uavpos_servo_loop(const uavpos_reference *in,
                  or_rigid_body_state *reference,
                  const genom_context self)
{
  const or_rigid_body_state *in_data;

  if (in->read(self)) return uavpos_e_input(self);
  in_data = in->data(self);
  if (!in_data) return uavpos_e_input(self);

  /* check if timestamps have changed */
  if (reference->ts.nsec == in_data->ts.nsec &&
      reference->ts.sec == in_data->ts.sec) return uavpos_pause_start;

  /* keep old value for missing fields for uavpos_reference_check */
#define copy_if_present(dst, src, f)                                    \
  do {                                                                  \
    dst->f._present = src->f._present;                                  \
    if (src->f._present) dst->f._value = src->f._value;                 \
  } while(0)

  copy_if_present(reference, in_data, pos);
  copy_if_present(reference, in_data, att);
  copy_if_present(reference, in_data, vel);
  copy_if_present(reference, in_data, avel);
  copy_if_present(reference, in_data, acc);
  copy_if_present(reference, in_data, aacc);
  copy_if_present(reference, in_data, jerk);
  copy_if_present(reference, in_data, snap);
#undef copy_if_present

  reference->ts = in_data->ts;
  reference->intrinsic = in_data->intrinsic;
  return uavpos_pause_start;
}

/** Codel uavpos_servo_stop of activity servo.
 *
 * Triggered by uavpos_stop.
 * Yields to uavpos_ether.
 * Throws uavpos_e_input.
 */
genom_event
uavpos_servo_stop(or_rigid_body_state *reference,
                  const genom_context self)
{
  (void)self; /* -Wunused-parameter */

  reference->intrinsic = false;

  reference->pos._present = true;
  reference->att._present = true;

  reference->vel._present = true;
  reference->vel._value.vx = 0.;
  reference->vel._value.vy = 0.;
  reference->vel._value.vz = 0.;

  reference->avel._present = true;
  reference->avel._value.wx = 0.;
  reference->avel._value.wy = 0.;
  reference->avel._value.wz = 0.;

  reference->acc._present = true;
  reference->acc._value.ax = 0.;
  reference->acc._value.ay = 0.;
  reference->acc._value.az = 0.;

  reference->aacc._present = false;
  reference->jerk._present = false;
  reference->snap._present = false;

  return uavpos_ether;
}


/* --- Activity set_current_position ------------------------------------ */

/** Codel uavpos_set_current_position of activity set_current_position.
 *
 * Triggered by uavpos_start.
 * Yields to uavpos_ether.
 * Throws uavpos_e_input.
 */
genom_event
uavpos_set_current_position(const uavpos_state *state,
                            or_rigid_body_state *reference,
                            const genom_context self)
{
  const or_pose_estimator_state *state_data;
  double qw, qx, qy, qz;
  double yaw;

  if (state->read(self)) return uavpos_e_input(self);
  state_data = state->data(self);
  if (!state_data) return uavpos_e_input(self);
  if (!state_data->pos._present) return uavpos_e_input(self);
  if (!state_data->att._present) return uavpos_e_input(self);

  qw = state_data->att._value.qw;
  qx = state_data->att._value.qx;
  qy = state_data->att._value.qy;
  qz = state_data->att._value.qz;
  yaw = atan2(2 * (qw*qz + qx*qy), 1 - 2 * (qy*qy + qz*qz));

  reference->ts = state_data->ts;
  reference->intrinsic = state_data->intrinsic;

  reference->pos._present = true;
  reference->pos._value.x = state_data->pos._value.x;
  reference->pos._value.y = state_data->pos._value.y;
  reference->pos._value.z = state_data->pos._value.z;

  reference->att._present = true;
  reference->att._value.qw = cos(yaw/2.);
  reference->att._value.qx = 0.;
  reference->att._value.qy = 0.;
  reference->att._value.qz = sin(yaw/2.);

  reference->vel._present = true;
  reference->vel._value.vx = 0.;
  reference->vel._value.vy = 0.;
  reference->vel._value.vz = 0.;

  reference->avel._present = true;
  reference->avel._value.wx = 0.;
  reference->avel._value.wy = 0.;
  reference->avel._value.wz = 0.;

  reference->acc._present = true;
  reference->acc._value.ax = 0.;
  reference->acc._value.ay = 0.;
  reference->acc._value.az = 0.;

  reference->aacc._present = false;
  reference->jerk._present = false;
  reference->snap._present = false;

  return uavpos_ether;
}
